import numpy

from purrfectmeow.meow.chaus import SimilarityResult
from purrfectmeow.meow.felis import Document

from .cosine import CosineSim


class KhaoManee:
    @classmethod
    def searching(
        cls,
        query_embed: numpy.ndarray,
        sentence_embed: numpy.ndarray | list[numpy.ndarray],
        documents: list[Document],
        top_k: int,
    ) -> list[SimilarityResult]:
        return CosineSim.vector_search(query_embed, sentence_embed, documents, top_k)



import time

import numpy

from purrfectmeow.meow.chaus import SimilarityResult
from purrfectmeow.meow.felis import Document
from purrfectmeow.meow.kitty import kitty_logger


class CosineSim:
    _logger = kitty_logger(__name__)

    @classmethod
    def vector_search(
        cls,
        embed_query: numpy.ndarray,
        embed_sentence: numpy.ndarray | list[numpy.ndarray],
        documents: list[Document],
        top_k: int,
    ) -> list[SimilarityResult]:
        cls._logger.debug("Initializing vector search")
        start = time.time()
        try:
            from sklearn.metrics.pairwise import cosine_similarity

            sims = cosine_similarity([embed_query], embed_sentence)[0]
            top_indices = numpy.argsort(sims)[::-1][:top_k]

            results: list[SimilarityResult] = [
                SimilarityResult(score=float(sims[i]), document=documents[i]) for i in top_indices
            ]

            return results
        except Exception as e:
            cls._logger.exception(f"Failed to initialize vector search: {e}")
            raise
        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Vector search completed in {elapsed:.2f} seconds.")



import time

from langchain_text_splitters import TokenTextSplitter

from purrfectmeow.meow.kitty import kitty_logger


class TokenSplit:
    _logger = kitty_logger(__name__)

    _OPENAI_EMBED_MODEL = {"text-embedding-ada-002", "text-embedding-3-small", "text-embedding-3-large"}
    _OPENAI_HF_MODEL = {"Xenova/text-embedding-ada-002"}
    _HF_MODEL_DIR = ".cache/huggingface/hub/"

    @classmethod
    def splitter(cls, model_name: str, chunk_size: int, chunk_overlap: int) -> TokenTextSplitter:
        cls._logger.debug("Initializing token splitter")
        start = time.time()

        try:
            cls._logger.debug(f"Using OpenAI model tokenizer: {model_name}")
            if model_name in cls._OPENAI_EMBED_MODEL:
                splitter = TokenTextSplitter.from_tiktoken_encoder(
                    model_name=model_name, chunk_size=chunk_size, chunk_overlap=chunk_overlap
                )
            else:
                cls._logger.debug(f"Using HuggingFace tokenizer: {model_name}")
                from transformers import AutoTokenizer, GPT2TokenizerFast

                if model_name in cls._OPENAI_HF_MODEL:
                    tokenizer = GPT2TokenizerFast.from_pretrained(model_name, cache_dir=cls._HF_MODEL_DIR)
                else:
                    tokenizer = AutoTokenizer.from_pretrained(model_name, cache_dir=cls._HF_MODEL_DIR)
                splitter = TokenTextSplitter.from_huggingface_tokenizer(
                    tokenizer=tokenizer, chunk_size=chunk_size, chunk_overlap=chunk_overlap
                )

            cls._logger.debug("Token splitter successfully initialized.")
            return splitter

        except Exception as e:
            cls._logger.exception(f"Failed to initialize token splitter: {e}")
            raise

        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Token splitting completed in {elapsed:.2f} seconds.")



import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path


class LevelBasedFormatter(logging.Formatter):
    def __init__(self, default_fmt: str, info_fmt: str, datefmt: str | None = None) -> None:
        super().__init__(datefmt=datefmt)
        self.default_fmt: logging.Formatter = logging.Formatter(default_fmt, datefmt)
        self.info_fmt: logging.Formatter = logging.Formatter(info_fmt, datefmt)

    def format(self, record: logging.LogRecord) -> str:
        if record.levelno == logging.INFO:
            return self.info_fmt.format(record)
        return self.default_fmt.format(record)


def kitty_logger(name: str, log_file: str = "kitty.log", log_level: str = "INFO") -> logging.Logger:
    """
    Sets up a logger with console and rotating file handlers.

    Args:
        name (str): Name of the logger (usually __name__ of the calling module).
        log_file (str): Path to the log file. Defaults to 'kitty.log'.
        log_level (str): Logging level ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'). Defaults to 'INFO'.

    Returns:
        logging.Logger: Configured logger instance.
    """
    logger = logging.getLogger(name)
    logger.setLevel(getattr(logging, log_level.upper(), logging.INFO))

    if not logger.handlers:
        default_fmt = "PurrfectKit | %(asctime)s [%(levelname)s] %(name)s:%(lineno)d - %(message)s"
        info_fmt = "PurrfectKit | %(asctime)s [%(levelname)s] - %(message)s"
        datefmt = "%Y-%m-%d %H:%M:%S"

        formatter = LevelBasedFormatter(default_fmt, info_fmt, datefmt)

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

        log_dir = Path(".cache/logs")
        log_dir.mkdir(parents=True, exist_ok=True)
        log_path = log_dir / log_file

        file_handler = RotatingFileHandler(log_path, maxBytes=5 * 1024 * 1024, backupCount=3)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger



from io import BytesIO
from typing import Any, TypedDict


class FileMetadata(TypedDict, total=False):
    file_name: str
    file_size: int
    file_created_date: str
    file_modified_date: str
    file_extension: str
    file_type: str
    description: str
    total_pages: int | str
    file_md5: str


class Document:
    def __init__(self, page_content: str, metadata: dict[str, Any]) -> None:
        self.page_content = page_content
        self.metadata = metadata or {}

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(page_content={self.page_content!r}, metadata={self.metadata!r})"

    def __getitem__(self, key: str) -> Any:
        if key == "page_content":
            return self.page_content
        elif key == "metadata":
            return self.metadata
        else:
            raise KeyError(f"{key} is not a valid key. Use 'page_content' or 'metadata'.")

    def to_dict(self) -> dict[str, Any]:
        return {"page_content": self.page_content, "metadata": self.metadata}


class DocTemplate:
    @staticmethod
    def create_template(chunks: list[str], metadata: dict[str, Any]) -> list[Document]:
        if not isinstance(chunks, list):
            raise TypeError(f"Expected 'chunks' to be a list, but got {type(chunks).__name__}.")

        if not isinstance(metadata, dict):
            raise TypeError(f"Expected 'metadata' to be a dict, but got {type(metadata).__name__}.")

        if not all(isinstance(c, str) for c in chunks):
            raise ValueError("All elements in 'chunks' must be strings.")

        docs = []
        chunk_hashes = []

        import hashlib
        import uuid

        for _, chunk in enumerate(chunks):
            hash_val = hashlib.md5(chunk.encode()).hexdigest()
            chunk_hashes.append(hash_val)

        for idx, chunk in enumerate(chunks):
            chunk_number = idx + 1
            chunk_id = uuid.uuid4().hex
            chunk_hash = chunk_hashes[idx]
            prev_hash = chunk_hashes[idx - 1] if idx > 0 else None
            next_hash = chunk_hashes[idx + 1] if idx < len(chunks) - 1 else None
            chunk_size = len(chunk)

            chunk_info = {
                "chunk_number": chunk_number,
                "chunk_id": chunk_id,
                "chunk_hash": chunk_hash,
                "previous_chunk_hash": prev_hash,
                "next_chunk_hash": next_hash,
                "chunk_size": chunk_size,
            }

            doc_metadata = {"chunk_info": chunk_info, "source_info": metadata}

            doc = Document(page_content=chunk, metadata=doc_metadata)
            docs.append(doc)

        return docs


class MetaFile:
    @staticmethod
    def get_metadata(file: str | BytesIO, **kwargs: Any) -> FileMetadata:
        if isinstance(file, bytes):
            file = BytesIO(file)

        if isinstance(file, BytesIO):
            import os

            os.makedirs(".cache/tmp", exist_ok=True)
            file_name = kwargs.get("file_name")

            if not file_name:
                raise ValueError("file_name must be provided when using BytesIO.")

            file_path = os.path.join(".cache/tmp", file_name)
            with open(file_path, "wb") as f:
                f.write(file.getvalue())

            try:
                return MetaFile._get_metadata_from_path(file_path)
            finally:
                os.remove(file_path)

        elif isinstance(file, str):
            return MetaFile._get_metadata_from_path(file)

        else:
            raise TypeError(f"Unsupported file type: {type(file).__name__}. Expected str, bytes, or BytesIO.")

    @staticmethod
    def _get_metadata_from_path(file_path: str) -> FileMetadata:
        metadata: FileMetadata = {}

        import hashlib
        import os
        import re
        import subprocess
        import time

        import magic

        try:
            if not os.path.exists(file_path):
                raise FileNotFoundError(f"File {file_path} does not exist")

            stats = os.stat(file_path)
            metadata["file_name"] = os.path.basename(file_path)
            metadata["file_size"] = stats.st_size
            metadata["file_created_date"] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(stats.st_ctime))
            metadata["file_modified_date"] = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(stats.st_mtime))
            metadata["file_extension"] = os.path.splitext(file_path)[1] or "none"

            try:
                mime = magic.Magic(mime=True)
                metadata["file_type"] = mime.from_file(file_path)
                metadata["description"] = magic.from_file(file_path)
            except Exception as e:
                metadata["file_type"] = "unknown"
                metadata["description"] = f"Could not determine file type: {str(e)}"

            if metadata["file_type"].startswith("image/"):
                metadata["total_pages"] = 1
            elif metadata["file_type"].startswith("application/pdf"):
                try:
                    result = subprocess.run(["pdfinfo", file_path], stdout=subprocess.PIPE, text=True, check=True)
                    pages_match = re.search(r"Pages:\s*(\d+)", result.stdout)
                    if pages_match:
                        metadata["total_pages"] = int(pages_match.group(1))
                    else:
                        metadata["total_pages"] = "Unknown (could not parse page count)"
                except (subprocess.CalledProcessError, FileNotFoundError):
                    metadata["total_pages"] = "Unknown (pdfinfo not installed or failed)"
            else:
                metadata["total_pages"] = 1

            with open(file_path, "rb") as f:
                hash_md5 = hashlib.md5()
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
                metadata["file_md5"] = hash_md5.hexdigest()

            return metadata

        except Exception as e:
            raise RuntimeError(f"Failed to extract metadata: {e}") from e



import time
from collections.abc import Callable
from typing import Any

from purrfectmeow.meow.kitty import kitty_logger


class Markdown:
    _logger = kitty_logger(__name__)

    @classmethod
    def _convert(cls, file_path: str, converter: Callable[[str], Any], extractor: Callable[[Any], str]) -> str:
        cls._logger.debug(f"Starting conversion for '{file_path}'")
        start = time.time()
        try:
            raw_content: Any = converter(file_path)
            result: str = extractor(raw_content)

            cls._logger.debug(f"Succesfully converted '{file_path}'")

            return result

        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Conversion time spent '{elapsed:.2f}' seconds.")

    @classmethod
    def markitdown_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using MarkItDown for Conversion")

        from markitdown import MarkItDown

        mid = MarkItDown()

        return cls._convert(file_path, lambda path: mid.convert(path), lambda content: content.text_content)

    @classmethod
    def docling_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using Docling for Conversion")

        from docling.document_converter import DocumentConverter

        dcl = DocumentConverter()

        return cls._convert(
            file_path, lambda path: dcl.convert(path).document, lambda content: content.export_to_markdown()
        )

    @classmethod
    def pymupdf4llm_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using PyMuPDF4LLM for Conversion")
        cls._logger.debug(f"Starting conversion for '{file_path}'")
        start = time.time()

        import pymupdf4llm

        try:
            res: str = pymupdf4llm.to_markdown(file_path)
            cls._logger.debug(f"Succesfully converted '{file_path}'")

            return res
        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Conversion time spent '{elapsed:.2f}' seconds.")



import time
from collections.abc import Callable
from typing import Any

from purrfectmeow.meow.kitty import kitty_logger


class Simple:
    _logger = kitty_logger(__name__)

    @classmethod
    def _convert(cls, file_path: str, converter: Callable[[str], Any]) -> str | Any:
        cls._logger.debug(f"Starting conversion for '{file_path}'")
        start = time.time()

        try:
            res = converter(file_path)

            cls._logger.debug(f"Successfully converted '{file_path}'")
            return res

        finally:
            elasped = time.time() - start
            cls._logger.debug(f"Conversion time spent '{elasped:.2f}' seconds.")

    @classmethod
    def encoding_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using Encoding for Conversion")

        def reader(file_path: str) -> str:
            with open(file_path, encoding="utf-8") as f:
                return f.read()

        return cls._convert(file_path, lambda file_path: reader(file_path))

    @classmethod
    def pymupdf_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using PyMuPDF for Conversion")

        def reader(file_path: str) -> str:
            import pymupdf

            if file_path.endswith((".txt", ".md", ".json", ".html", ".xml")):
                return "".join(page.get_text() for page in pymupdf.open(file_path, filetype="txt"))
            else:
                return "".join(page.get_text() for page in pymupdf.open(file_path))

        return cls._convert(file_path, lambda file_path: reader(file_path))

    @classmethod
    def pandas_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using Pandas for Conversion")

        def reader(file_path: str) -> Any:
            import pandas

            if file_path.endswith((".xls", ".xlsx")):
                df_x: pandas.DataFrame = pandas.read_excel(file_path)
                return df_x.to_string(index=False)
            elif file_path.endswith(".csv"):
                df_c: pandas.DataFrame = pandas.read_csv(file_path)
                return df_c.to_string(index=False)
            elif file_path.endswith(".json"):
                df_j: pandas.DataFrame = pandas.read_json(file_path)
                return df_j.to_string(index=False)
            elif file_path.endswith(".html"):
                df_h: list[pandas.DataFrame] = pandas.read_html(file_path)
                return "".join(df.to_string(index=False) for df in df_h)
            elif file_path.endswith(".xml"):
                df_m: pandas.DataFrame = pandas.read_xml(file_path)
                return df_m.to_string(index=False)
            else:
                return ""

        return cls._convert(file_path, lambda file_path: reader(file_path))



import os
import time
from typing import Any

from purrfectmeow.meow.kitty import kitty_logger


class OpenAI:
    _logger = kitty_logger(__name__)

    MAX_TOKENS_OPENAI = 300_000

    @classmethod
    def model_encode(
        cls, sentence: str | list[str], model_name: str, base_url: str, api_key: str, **kwargs: Any
    ) -> dict[str, Any]:
        cls._logger.debug("Initializing OpenAI encode")
        start = time.time()
        try:
            if api_key:
                key = api_key
            else:
                key = os.environ.get("OPENAI_API_KEY") or ""

            import tiktoken
            from openai import OpenAI

            tokenizer = tiktoken.get_encoding("cl100k_base")
            client = OpenAI(base_url=base_url, api_key=key)

            all_embeddings = []
            total_tokens = 0
            current_batch: list[str] = []
            current_tokens = 0
            batch_count = 0

            def submit_batch(batch: list[str]) -> None:
                nonlocal total_tokens, all_embeddings, batch_count
                batch_count += 1
                cls._logger.debug(
                    f"Submitting OpenAI batch {batch_count} with {len(batch)} texts ({current_tokens} tokens)"
                )
                try:
                    response = client.embeddings.create(input=batch, model=model_name)
                    all_embeddings.extend([item.embedding for item in response.data])
                    total_tokens += response.usage.total_tokens
                    cls._logger.debug(
                        f"OpenAI batch {batch_count} succeeded - tokens used: {response.usage.total_tokens}"
                    )
                except Exception as e:
                    cls._logger.error(f"OpenAI model '{model_name}' failed to create embeddings: {e}")
                    raise

            for text in sentence:
                token_count = len(tokenizer.encode(text))
                if current_tokens + token_count > cls.MAX_TOKENS_OPENAI:
                    cls._logger.debug(
                        f"Token limit reached ({current_tokens}/{cls.MAX_TOKENS_OPENAI}), submitting batch"
                    )
                    submit_batch(current_batch)
                    current_batch = []
                    current_tokens = 0

                current_batch.append(text)
                current_tokens += token_count

            if current_batch:
                submit_batch(current_batch)

            cls._logger.debug("Openai encode successfully initialized.")
            return {"embeddings": all_embeddings, "total_tokens": total_tokens}

        except Exception as e:
            cls._logger.exception(f"Failed to initialize OpenAI encode: {e}")
            raise

        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"OpenAI encode completed in {elapsed:.2f} seconds.")



from .tc01_spl import Suphalak
from .tc02_mlt import Malet
from .tc03_wcm import WichienMaat
from .tc04_kmn import KhaoManee
from .tc05_knj import Kornja

__all__ = [
    "Suphalak",
    "Malet",
    "WichienMaat",
    "KhaoManee",
    "Kornja",
]

__version__ = "0.2.7"



import time
from collections.abc import Callable
from typing import Any

from purrfectmeow.meow.kitty import kitty_logger


class Ocr:
    _logger = kitty_logger(__name__)
    _image_type = [
        ".apng",
        ".png",
        ".avif",
        ".gif",
        ".jpg",
        ".jpeg",
        ".jfif",
        ".pjpeg",
        ".pjp",
        ".png",
        ".svg",
        ".webp",
        ".bmp",
        ".ico",
        ".cur",
        ".tif",
        ".tiff",
    ]

    @classmethod
    def _convert(cls, file_path: str, converter: Callable[[str], Any]) -> str:
        cls._logger.debug(f"Starting conversion for '{file_path}'")
        start = time.time()

        try:
            content = []
            match file_path.lower():
                case path if path.endswith(".pdf"):
                    from pdf2image import convert_from_path

                    images = convert_from_path(file_path, fmt="png")
                    for idx, image in enumerate(images):
                        try:
                            text = converter(image)
                            cls._logger.debug(f"Text: {text}")
                            content.append(text)
                            cls._logger.debug(f"Page {idx + 1} processed")
                        except Exception as e:
                            cls._logger.exception(f"Page {idx + 1} failed: {e}")
                            raise
                case path if path.endswith(tuple(cls._image_type)):
                    from PIL import Image

                    image = Image.open(file_path)
                    try:
                        text = converter(image)
                        cls._logger.debug(f"Text: {text}")
                        content.append(text)
                        cls._logger.debug("Page 1 processed")
                    except Exception as e:
                        cls._logger.debug(f"Page 1 failed: {e}")
                        raise

            cls._logger.debug(f"Successfully converted '{file_path}'")
            return "\n".join(content)

        finally:
            elasped = time.time() - start
            cls._logger.debug(f"Conversion time spent '{elasped:.2f}' seconds.")

    @classmethod
    def pytesseract_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using PyTesseract for Conversion")

        def converter(image: str) -> Any:
            import pytesseract

            return pytesseract.image_to_string(image, lang="tha+eng")

        return cls._convert(file_path, converter)

    @classmethod
    def easyocr_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using EasyOCR for Conversion")

        def converter(image: str) -> Any:
            import easyocr
            import numpy

            reader = easyocr.Reader(["th", "en"], gpu=False)
            res = reader.readtext(numpy.array(image))
            return "\n".join(text for _, text, _ in res)

        return cls._convert(file_path, converter)

    @classmethod
    def suryaocr_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using SuryaOCR for Conversion")

        def converter(image: str) -> Any:
            from surya.detection import DetectionPredictor
            from surya.recognition import RecognitionPredictor

            rec_pred = RecognitionPredictor()
            det_pred = DetectionPredictor()

            prediction = rec_pred(
                [image],
                det_predictor=det_pred,
                detection_batch_size=1,
                recognition_batch_size=1,
            )
            return "\n".join(line.text for line in prediction[0].text_lines)

        return cls._convert(file_path, converter)

    @classmethod
    def doctr_convert(cls, file_path: str) -> str:
        cls._logger.debug("Using docTR for Conversion")

        def converter(image: Any) -> Any:
            import os
            import tempfile

            from doctr.io import DocumentFile
            from doctr.models import ocr_predictor

            with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as tmp:
                image.save(tmp.name)
                tmp_path = tmp.name

            model = ocr_predictor(pretrained=True)
            doc = DocumentFile.from_images(tmp_path)
            result = model(doc)
            data = result.export()
            combined_text = "\n".join(
                word["value"]
                for page in data["pages"]
                for block in page.get("blocks", [])
                for line in block.get("lines", [])
                for word in line.get("words", [])
                if "value" in word
            )
            if os.path.exists(tmp_path):
                os.remove(tmp_path)
            return combined_text

        return cls._convert(file_path, converter)

    @classmethod
    def typhoonocr_convert(cls, file_path: str, **kwargs: Any) -> str:
        cls._logger.debug("Using Typhoon OCR for Conversion")

        def converter(image: Any) -> Any:
            import os
            import re
            import tempfile

            from typhoon_ocr import ocr_document

            base_url = kwargs.get("base_url") or os.getenv("OLLAMA_SERVER") or "http://localhost:11434/v1"
            if not base_url.endswith("/v1"):
                base_url += "/v1"
            api_key = kwargs.get("api_key") or "ollama"
            model = kwargs.get("model") or kwargs.get("model_name") or "scb10x/typhoon-ocr1.5-3b"

            with tempfile.NamedTemporaryFile(suffix=".png", delete=False) as tmp:
                image.save(tmp.name)
                tmp_path = tmp.name

            text = ocr_document(tmp_path, base_url=base_url, api_key=api_key, model=model)

            if model.startswith("gemma3"):
                text = re.sub(r"^```[a-zA-Z0-9_]*\n|```$", "", text).strip()

            os.remove(tmp_path)

            return text

        return cls._convert(file_path, converter)



from typing import Any, BinaryIO

from .markdown import Markdown
from .ocr import Ocr
from .simple import Simple


class Suphalak:
    tmp_dir = ".cache/tmp"
    DEFAULT_LOADER = "PYMUPDF4LLM"

    _LOADERS: dict[str, dict[str, Any]] = {
        "MARKITDOWN": {
            "func": Markdown.markitdown_convert,
            "ext": ("csv", "docx", "md", "pdf", "pptx", "txt", "xls", "xlsx"),
        },
        "DOCLING": {
            "func": Markdown.docling_convert,
            "ext": ("csv", "docx", "jpg", "jpeg", "md", "pdf", "png", "pptx", "xlsx"),
        },
        "PYMUPDF4LLM": {
            "func": Markdown.pymupdf4llm_convert,
            "ext": ("docx", "pdf", "pptx", "txt", "xlsx"),
        },
        "PYTESSERACT": {
            "func": Ocr.pytesseract_convert,
            "ext": ("gif", "jpg", "jpeg", "pdf", "png"),
        },
        "EASYOCR": {
            "func": Ocr.easyocr_convert,
            "ext": ("gif", "jpg", "jpeg", "pdf", "png"),
        },
        "SURYAOCR": {
            "func": Ocr.suryaocr_convert,
            "ext": ("gif", "jpg", "jpeg", "pdf", "png"),
        },
        "DOCTR": {
            "func": Ocr.doctr_convert,
            "ext": ("gif", "jpg", "jpeg", "pdf", "png"),
        },
        "TYPHOONOCR": {
            "func": Ocr.typhoonocr_convert,
            "ext": ("gif", "jpg", "jpeg", "pdf", "png"),
        },
        "PYMUPDF": {
            "func": Simple.pymupdf_convert,
            "ext": ("docx", "md", "pdf", "pptx", "xlsx"),
        },
        "PANDAS": {
            "func": Simple.pandas_convert,
            "ext": ("csv", "xls", "xlsx"),
        },
        "ENCODING": {
            "func": Simple.encoding_convert,
            "ext": ("csv", "md", "txt"),
        },
    }

    @classmethod
    def _detect_loader(cls, file_ext: str) -> str:
        priority = [
            ("PANDAS", ("csv", "xls")),
            ("PYTESSERACT", ("jpg", "jpeg", "png", "gif")),
            ("PYMUPDF", ("pdf", "md")),
            ("PYMUPDF4LLM", ("txt", "xlsx", "pptx", "docx")),
        ]

        for loader, extensions in priority:
            if file_ext in extensions:
                return loader

        return cls.DEFAULT_LOADER

    @classmethod
    def reading(cls, file: BinaryIO, file_name: str, loader: str | None = None, **kwargs: Any) -> str:
        import os

        file_ext = file_name.split(".")[-1].lower()

        if not loader:
            loader = cls._detect_loader(file_ext)

        if loader not in cls._LOADERS:
            raise ValueError(f"Unsupported loader: '{loader}'")

        loader_conf = cls._LOADERS[loader]
        supported_ext = loader_conf["ext"]

        if file_ext not in supported_ext:
            raise TypeError(f"'{file_ext}' is not supported for '{loader}' loader.")

        os.makedirs(cls.tmp_dir, exist_ok=True)
        file_path = os.path.join(cls.tmp_dir, file_name)

        try:
            text: str
            with open(file_path, "wb") as f:
                f.write(file.read())

            text = loader_conf["func"](file_path, **kwargs)

            if (
                file_ext == "pdf"
                and (not text or not str(text).strip())
                and loader not in ("PYTESSERACT", "EASYOCR", "SURYAOCR", "DOCTR", "TYPHOONOCR")
            ):
                ocr_loader = cls._LOADERS["PYTESSERACT"]
                text = ocr_loader["func"](file_path, **kwargs)

            return text

        finally:
            if os.path.exists(file_path):
                os.remove(file_path)



import os
import time
from typing import Any

from purrfectmeow.meow.kitty import kitty_logger


class Ollama:
    _logger = kitty_logger(__name__)

    @classmethod
    def model_encode(
        cls, sentence: str | list[str], model_name: str, ollama_url: str | None, **kwargs: Any
    ) -> dict[str, Any]:
        cls._logger.debug("Initializing ollama encode")
        start = time.time()
        try:
            from ollama import Client

            if ollama_url:
                host = ollama_url

            else:
                host = os.environ.get("OLLAMA_HOST") or "http://localhost:11434"

            client = Client(host=host)
            embed = client.embed(model=model_name, input=sentence)

            cls._logger.debug("Ollama encode successfully initialized.")
            return {"embeddings": embed.get("embeddings"), "total_tokens": embed.get("prompt_eval_count")}

        except Exception as e:
            cls._logger.exception(f"Failed to initialize Ollama encode: {e}")
            raise

        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Ollama encode completed in {elapsed:.2f} seconds.")



import time
from typing import Any

from sentence_transformers import SentenceTransformer

from purrfectmeow.meow.kitty import kitty_logger


class HuggingFace:
    _logger = kitty_logger(__name__)
    _MODEL_CACHE: dict[str, SentenceTransformer] = {}

    @classmethod
    def model_encode(
        cls,
        sentence: str | list[str],
        model_name: str,
    ) -> dict[str, Any]:
        cls._logger.debug("Initializing huggingface encode")
        start = time.time()
        try:
            if model_name not in cls._MODEL_CACHE:
                cls._logger.debug(f"Loading model: {model_name}")
                cls._MODEL_CACHE[model_name] = SentenceTransformer(model_name)
            else:
                cls._logger.debug(f"Using cached model: {model_name}")

            model = cls._MODEL_CACHE[model_name]

            cls._logger.debug("Encoding sentence(s)")
            embeddings = model.encode(sentence)

            cls._logger.debug("Encoding completed successfully")

            # Count tokens using the model's tokenizer
            tokenizer = model.tokenizer
            input_list = sentence if isinstance(sentence, list) else [sentence]
            total_tokens = sum(len(tokenizer.encode(text)) for text in input_list)

            if isinstance(sentence, str):
                return {"embeddings": [embeddings.tolist()], "total_tokens": total_tokens}

            return {"embeddings": embeddings.tolist(), "total_tokens": total_tokens}

        except Exception as e:
            cls._logger.exception(f"Failed to initialize Huggingface encode: {e}")
            raise

        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Huggingface encode completed in {elapsed:.2f} seconds.")



from typing import Any, Literal

from langchain_text_splitters import TokenTextSplitter

from .separate import SeparateSplit
from .token import TokenSplit


class Malet:
    DEFAULT_MODEL_NAME = "sentence-transformers/paraphrase-multilingual-MiniLM-L12-v2"
    DEFAULT_CHUNK_SIZE = 500
    DEFAULT_CHUNK_OVERLAP = 0
    DEFAULT_CHUNK_SEPARATOR = "\n\n"

    @staticmethod
    def _get_kwarg(kwargs: dict[str, Any], keys: list[str], default: Any = None) -> Any:
        for key in keys:
            if key in kwargs:
                return kwargs[key]
        return default

    @classmethod
    def chunking(
        cls, text: str, chunk_method: Literal["token", "separate"] | None = "token", **kwargs: Any
    ) -> TokenTextSplitter | SeparateSplit.CharacterSeparator:
        match chunk_method:
            case "token":
                model_name = cls._get_kwarg(kwargs, ["model_name", "ModelName", "modelName"], cls.DEFAULT_MODEL_NAME)
                chunk_size = cls._get_kwarg(kwargs, ["chunk_size", "ChunkSize", "chunkSize"], cls.DEFAULT_CHUNK_SIZE)
                chunk_overlap = cls._get_kwarg(
                    kwargs, ["chunk_overlap", "ChunkOverlap", "chunkOverlap"], cls.DEFAULT_CHUNK_OVERLAP
                )

                method = TokenSplit.splitter(model_name, chunk_size, chunk_overlap)

            case "separate":
                chunk_separator = cls._get_kwarg(
                    kwargs, ["chunk_separator", "ChunkSeparator", "chunkSeparator"], cls.DEFAULT_CHUNK_SEPARATOR
                )

                method = SeparateSplit.splitter(chunk_separator)

        return method.split_text(text)



from __future__ import annotations

import time

from purrfectmeow.meow.kitty import kitty_logger


class SeparateSplit:
    _logger = kitty_logger(__name__)

    @classmethod
    def splitter(cls, chunk_separator: str) -> CharacterSeparator:
        cls._logger.debug("Initializing separate splitter")
        start = time.time()

        try:
            splitter = cls.CharacterSeparator(chunk_separator)

            cls._logger.debug("Separator splitter successfully initialized.")
            return splitter
        except Exception as e:
            cls._logger.exception(f"Failed to initialize separate splitter: {e}")
            raise
        finally:
            elapsed = time.time() - start
            cls._logger.debug(f"Separate splitting completed in {elapsed:.2f} seconds.")

    class CharacterSeparator:
        def __init__(self, separator: str):
            self.separator = separator

        def split_text(self, text: str) -> list[str]:
            chunks = [chunk + self.separator for chunk in text.split(self.separator)]
            chunks[-1] = chunks[-1].rstrip(self.separator)
            return chunks



from typing import Any

import yaml

from .huggingface import HuggingFace
from .ollama import Ollama
from .openai import OpenAI


class WichienMaat:
    @classmethod
    def embedding(cls, sentence: str | list[str], model_name: str | None = None, **kwargs: Any) -> dict[str, Any]:
        ollama_url = kwargs.get("ollama_url") or kwargs.get("ollama_server") or kwargs.get("ollama_host") or ""
        base_url = kwargs.get("openai_base_url") or kwargs.get("base_url") or ""
        api_key = kwargs.get("openai_api_key") or kwargs.get("api_key") or ""

        with open("purrfectmeow/meow/felidae/models.yaml") as f:
            _CFG = yaml.safe_load(f)

        hf_models = set(_CFG["huggingface"]["models"])
        hf_default = _CFG["huggingface"]["default"]

        ollama_models = set(_CFG["ollama"]["models"])
        ollama_default = _CFG["ollama"]["default"]

        openai_models = set(_CFG["openai"]["models"])
        openai_default = _CFG["openai"]["default"]

        match model_name:
            case _ if model_name is not None and model_name in hf_models:
                return HuggingFace.model_encode(sentence, model_name)

            case _ if model_name is not None and model_name in ollama_models:
                return Ollama.model_encode(sentence, model_name, ollama_url)

            case _ if model_name is not None and model_name in openai_models:
                if api_key:
                    return OpenAI.model_encode(sentence, model_name, base_url, api_key)
                elif ollama_url:
                    return Ollama.model_encode(sentence, model_name, ollama_url)
                else:
                    return HuggingFace.model_encode(sentence, model_name)

            case None:
                if api_key:
                    return OpenAI.model_encode(sentence, openai_default, base_url, api_key)

                elif ollama_url:
                    return Ollama.model_encode(sentence, ollama_default, ollama_url)

                return HuggingFace.model_encode(sentence, hf_default)

            case _:
                return HuggingFace.model_encode(sentence, model_name or hf_default)
